@@ -13,10 +13,52 @@ import pprint
 import enum
 import z3
 import time
+import fnmatch
+from functools import reduce
 from kmax.arch import Arch
 from kmax.common import get_kmax_constraints, unpickle_kmax_file
 from kmax.vcommon import getLogLevel, getLogger, run
 
+builtin_rewrite_mapping = {
+  "drivers/gpu/drm/amd/": "drivers/gpu/drm/amd/amdgpu/../",
+  "drivers/gpu/drm/amd/amdgpu/": "drivers/gpu/drm/amd/amdgpu/",  # this path is more specific, so the parent directory will not be rewritten for amdgpu
+  "virt/kvm/": "arch/x86/kvm/../../../virt/kvm", # note that virt/kvm also applies to arm
+}
+
+builtin_build_targets = {
+  "arch/arm/mach-omap1/*": "arch/arm/mach-omap1/",
+  "arch/arm/mach-pxa/*": "arch/arm/mach-pxa/",
+  "arch/arm/mach-s3c/*": "arch/arm/mach-s3c/",
+  "arch/arm/plat-omap/*": "arch/arm/plat-omap/",
+  "arch/x86/boot/compressed/*": "all",
+  "arch/x86/kvm/mmu/*": "arch/x86/kvm/",
+  "arch/x86/lib/*": "arch/x86/kvm/",
+  "virt/kvm/*": "arch/x86/kvm/",
+}
+
+def rewrite_directories(unit: str, rewrite_mapping: dict):
+  """Rewrite the longest path prefix from the rewrite mapping."""
+  # slow algorithm to find the longest matching string prefix from the keyset
+  # TODO: cache a trie of the prefix strings to find the longest match instead
+  candidates = [ s for s in rewrite_mapping.keys() if unit.startswith(s) ]
+  # two equivalently long candidates shouldn't be possible unless they
+  # are identical, since they are first filtered by being a prefix of
+  # unit
+  if len(candidates) > 0:
+    longest = reduce(lambda acc, key: key if len(str(key)) > len(str(acc)) else acc, candidates, "")
+    updated_unit = unit.replace(longest, rewrite_mapping[longest])
+    return updated_unit
+  else:
+    return unit
+
+def rewrite_build_target(unit: str, build_targets: dict):
+  """Rewrite the first matching build target pattern to build target mapping, or leave the original unit as is."""
+  for pattern, rewrite in build_targets.items():
+    if fnmatch.fnmatch(unit, pattern):
+      return rewrite
+  # when there is no rewrite, just return the original
+  return unit
+  
 # TODO(necip): automated testing
 
 def __parse_cb(cb_string_rep: str):
@@ -321,7 +363,40 @@ class Klocalizer:
       return constraints
   
   @staticmethod
-  def get_config_from_model(model: z3.Model, arch: Arch, set_tristate_m, allow_non_visibles: bool, logger=None) -> str:
+  def get_config_file_settings(config_file):
+    """Given a path to a Kconfig configuration file, parse and return a
+    mapping of non-Boolean (string, number, hex) config option names
+    to their original settings in the config file.  Excludes options
+    that are turned off.
+
+    Arguments:
+    config_file -- Path to a Kconfig config file.
+
+    """
+    assert os.path.isfile(config_file)
+    on_pattern = regex.compile("^(CONFIG_[A-Za-z0-9_]+)=[ym]")
+    on_pattern_nonbool = regex.compile("^(CONFIG_[A-Za-z0-9_]+)=.*$")
+    
+    # todo: don't allow invisible defaults to be turned off (get them from kclause), reduces size of constraints
+
+    config_settings = {}
+    with open(config_file, 'r') as approximate_fp:
+      lines = approximate_fp.readlines()
+      for line in lines:
+        line = line.strip()
+        on = on_pattern.match(line)
+        if on:
+          # exclude bool/tristate
+          pass
+        else:
+          on_nonbool = on_pattern_nonbool.match(line)
+          if on_nonbool:
+            config_settings[on_nonbool.group(1)] = line
+      
+      return config_settings
+  
+  @staticmethod
+  def get_config_from_model(model: z3.Model, arch: Arch, set_tristate_m, allow_non_visibles: bool, approximate_config=None, logger=None) -> str:
     """Given a z3_model, return a Kconfig config in string representation,
     which can be later dumped as a Kconfig config file and used against Kbuild
     build system.
@@ -335,6 +410,7 @@ class Klocalizer:
     set_tristate_m -- If set, set tristate symbols to `m` if on.
     allow_non_visibles -- Allow non-visible Kconfig configuration options to
     be set in the resulting config file.
+    approximate_config -- the original config file to approximate.  if set, will use this to set nonbool values instead of the default values.  otherwise, set to None (the default)
     logger -- Logger.
     
     """
@@ -358,6 +434,25 @@ class Klocalizer:
       kconfig_types = None
       kconfig_has_def_nonbool = None
 
+    if approximate_config is not None:
+      approximate_config_original = Klocalizer.get_config_file_settings(approximate_config)
+    else:
+      # make it an empty set so it will never contain config settings to use
+      approximate_config_original = {}
+    nonbool_defs = {}
+    for line in arch.get_kextract().splitlines():
+      if line.startswith("def_nonbool "):
+        instr, data = line.strip().split(" ", 1)
+        var, val_and_expr = data.split(" ", 1)
+        val, expr = val_and_expr.split("|", 1)
+        # TODO: this will use the last def if there are multiple ones.
+        # use constraint solver to pick the right one.
+        if val.startswith('"') and val.endswith('"'):
+          nonbool_defs[var] = val[1:-1] # strip the quotes
+        else:
+          # default value is from another variable, so it won't have a quoted value
+          # TODO: support assigning default values from other config options
+          pass
     for entry in model: # the model has some problem, we can't get the entry
       str_entry = str(entry)
       matches = token_pattern.match(str_entry)
@@ -383,11 +478,26 @@ class Klocalizer:
                     elif kconfig_types[str_entry] == "tristate":
                       write_to_content( "%s=%s\n" % (str_entry, "y" if not set_tristate_m else "m") )
                     elif kconfig_types[str_entry] == "string":
-                      write_to_content( "%s=\n" % (str_entry) )
+                      if str_entry in approximate_config_original.keys():
+                        write_to_content("%s\n" % (approximate_config_original[str_entry]))
+                      elif str_entry in nonbool_defs.keys():
+                        write_to_content("%s=\"%s\"\n" % (str_entry, nonbool_defs[str_entry]))
+                      else:
+                        write_to_content("%s=\n" % (str_entry) )
                     elif kconfig_types[str_entry] == "number":
-                      write_to_content( "%s=0\n" % (str_entry) )
+                      if str_entry in approximate_config_original.keys():
+                        write_to_content("%s\n" % (approximate_config_original[str_entry]))
+                      elif str_entry in nonbool_defs.keys():
+                        write_to_content("%s=%s\n" % (str_entry, nonbool_defs[str_entry]))
+                      else:
+                        write_to_content( "%s=0\n" % (str_entry) )
                     elif kconfig_types[str_entry] == "hex":
-                      write_to_content( "%s=0x0\n" % (str_entry) )
+                      if str_entry in approximate_config_original.keys():
+                        write_to_content("%s\n" % (approximate_config_original[str_entry]))
+                      elif str_entry in nonbool_defs.keys():
+                        write_to_content("%s=%s\n" % (str_entry, nonbool_defs[str_entry]))
+                      else:
+                        write_to_content("%s=0x0\n" % (str_entry) )
                     else:
                       assert(False)
                   break
@@ -400,11 +510,26 @@ class Klocalizer:
             elif kconfig_types[str_entry] == "tristate":
               write_to_content( "%s=%s\n" % (str_entry, "y" if not set_tristate_m else "m") )
             elif kconfig_types[str_entry] == "string":
-              write_to_content( "%s=\n" % (str_entry) )
+              if str_entry in approximate_config_original.keys():
+                write_to_content("%s\n" % (approximate_config_original[str_entry]))
+              elif str_entry in nonbool_defs.keys():
+                write_to_content("%s=\"%s\"\n" % (str_entry, nonbool_defs[str_entry]))
+              else:
+                write_to_content("%s=\n" % (str_entry) )
             elif kconfig_types[str_entry] == "number":
-              write_to_content( "%s=0\n" % (str_entry) )
+              if str_entry in approximate_config_original.keys():
+                write_to_content("%s\n" % (approximate_config_original[str_entry]))
+              elif str_entry in nonbool_defs.keys():
+                write_to_content("%s=%s\n" % (str_entry, nonbool_defs[str_entry]))
+              else:
+                write_to_content("%s=0\n" % (str_entry) )
             elif kconfig_types[str_entry] == "hex":
-              write_to_content( "%s=0x0\n" % (str_entry) )
+              if str_entry in approximate_config_original.keys():
+                write_to_content("%s\n" % (approximate_config_original[str_entry]))
+              elif str_entry in nonbool_defs.keys():
+                write_to_content("%s=%s\n" % (str_entry, nonbool_defs[str_entry]))
+              else:
+                write_to_content("%s=0x0\n" % (str_entry) )
             else:
               assert(False)
           else:
@@ -939,7 +1064,7 @@ class Klocalizer:
 
   @staticmethod
   def localize_config(unit_paths: list, output_config_path: str,
-    linux_ksrc: str, archs: list, logger):
+    linux_ksrc: str, archs: list, approximate_configs: dict, logger):
     """Localize a linux configuration file that compiles the given units.
 
     This is a simplified version of what klocalizer command line tool does:
@@ -963,10 +1088,11 @@ class Klocalizer:
     * output_config_path -- Path to non-existing output configuration file.
     * linux_ksrc -- Path to top linux source directory.
     * archs -- Architectures to try as a list of Arch instances.
+    * approximate_configs -- each arch's original config files to approximate.   used to get a config to pass through to get_config_from_model
     * logger -- Logger that implements common.VoidLogger interface.
     """
     try:
-      return Klocalizer.__localize_config(unit_paths, output_config_path, linux_ksrc, archs, logger)
+      return Klocalizer.__localize_config(unit_paths, output_config_path, linux_ksrc, archs, approximate_configs, logger)
     except Klocalizer.KmaxException as e:
       logger.debug("Exception in localize_config(): \"%s\"\n" % e.message)
       ret_status = Klocalizer.Ret_LocalizeConfig.ERROR_KMAX_EXCEPTION
@@ -983,7 +1109,7 @@ class Klocalizer:
 
   @staticmethod
   def __localize_config(unit_paths: list, output_config_path: str,
-    linux_ksrc: str, archs: list, logger):
+    linux_ksrc: str, archs: list, approximate_configs: dict, logger):
     """Helper to Klocalizer.localize_config(). May throw exception to be
     caught by Klocalizer.localize_config().
     """
@@ -1039,7 +1165,13 @@ class Klocalizer:
     logger.debug("Creating config file to compile the unit.\n")
     time_start = time.time()
     assert is_sat and (z3_model is not None)
-    config_content = Klocalizer.get_config_from_model(z3_model, arch, set_tristate_m=False, allow_non_visibles=False)
+    if arch.name in approximate_configs:
+      approximate_config = approximate_configs[arch.name]
+    elif "default" in approximate_configs:
+      approximate_config = approximate_configs["default"]
+    else:
+      approximate_config = None
+    config_content = Klocalizer.get_config_from_model(z3_model, arch, approximate_config=approximate_config, set_tristate_m=False, allow_non_visibles=False)
     assert config_content
     with open(output_config_path, "w") as f:
       f.write(config_content)
